//*****************************************************************************
// FILE..........: BTreeArrayBlock.hpp
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2013-06-21 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_BTREEARRAYBLOCK_HPP
#define FastCpp_BTREEARRAYBLOCK_HPP

//--------------------------------------------------------------- import std lib
//------------------------------------------------------------------------------
#ifndef                     FastCpp_BTREEARRAYBLOCKINTERFACE_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayBlockInterface.hpp>
#endif//                    FastCpp_BTREEARRAYBLOCKINTERFACE_HPP

#ifndef                     FastCpp_BTREEARRAYUPBLOCKINTERFACE_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayUpBlockInterface.hpp>
#endif//                    FastCpp_BTREEARRAYUPBLOCKINTERFACE_HPP
//------------------------------------------------------------------------------
namespace FastCpp
{
namespace Detail
{
//------------------------------------------------------------------------------
//| |       |                 |                 |                            |

template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
class BTreeArrayBlock : public BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
{
  public:
    typedef BTreeArrayBlockInterface  <T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayBlockInterfaceType;
    typedef BTreeArrayUpBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayUpBlockInterfaceType;
    typedef ElementReference          <T, UIndex, BIndex, BLOCK_SIZE, Allocator> ElementReferenceType;
    typedef BTreeArrayBlock                                                      ThisType;
  private:
    inline                    BTreeArrayBlock   (const BTreeArrayBlock& rObj);// = deleted;
    inline                    BTreeArrayBlock   (const BTreeArrayBlock& rObj, Allocator& rAllocator);
    inline                    BTreeArrayBlock   (
                                                  bool             bSecondHalf, // must be true
                                                  BTreeArrayBlock& rFullBlock,  // must be full
                                                  Allocator&       rAllocator
                                                );
    inline  void              operator =        (const BTreeArrayBlock& rObj);
  public:
    inline                    BTreeArrayBlock   (Allocator& rAllocator, BTreeArrayUpBlockInterfaceType* pUpNode = NULL);
    virtual void              Deallocate        (Allocator& rAllocator);
    virtual void              DestroyAndDeallocate
                                                (Allocator& rAllocator);
    virtual                  ~BTreeArrayBlock   ();
  public:
    inline  UIndex            GetUIndex         (BIndex nBIndex);
            ElementReferenceType
                              Insert            (const T& rElemt, ElementReferenceType& oReference);
            ElementReferenceType
                              Erase             (ElementReferenceType& oReference);
  //----------------------------------------------------BTreeArrayBlockInterface
  public:
    virtual BTreeArrayBlockInterfaceType*
                              Clone             (Allocator& rAllocator) const;
    virtual void              SetUpNode         (BTreeArrayUpBlockInterfaceType* pUpNode);
    virtual BTreeArrayBlockInterfaceType*
                              Insert            (
                                                  UIndex                        nUIndex,
                                                  const T&                      rElemt,
                                                  BTreeArrayBlockInterfaceType* pArrPrev,
                                                  BTreeArrayBlockInterfaceType* pNext,
                                                  Allocator&                    rAllocator
                                                );

    virtual T                 Replace           (UIndex nUIndex, const T& rBy);

    virtual UIndex            GetCount          ();
    virtual BIndex            GetItemsCount     () const;
    virtual UIndex            Depth             () const;
    virtual BTreeArrayBlockInterfaceType*
                              PushFront         (const T& rElemt, Allocator& rAllocator);
    virtual BTreeArrayBlockInterfaceType*
                              Add               (const T& rElemt, Allocator& rAllocator);
    virtual BTreeArrayBlockInterfaceType*
                              PlusOne           (Allocator& rAllocator);
    virtual T&                Get               (UIndex nUIndex);
    virtual T&                GetFirst          ();
    virtual T&                GetLast           ();

    virtual ElementReferenceType
                              Begin             ();
    virtual ElementReferenceType
                              Last              ();
            ElementReferenceType&
                              Prev              (ElementReferenceType& rElement);
            ElementReferenceType&
                              Next              (ElementReferenceType& rElement);
            ElementReferenceType&
                              Next              (
                                                  ElementReferenceType& rElement,
                                                  ptrdiff_t             nOffset
                                                );
    virtual ElementReferenceType
                              GetElementReference
                                                (UIndex nUIndex);
    virtual void              Constate          ();
    virtual void              ShrinkToFit       (Allocator& rAllocator);
    virtual UIndex            Delete            (
                                                  UIndex     nUIndex,
                                                  T&         rRet,
                                                  bool&      bSubBlockPacked,
                                                  Allocator& rAllocator
                                                );
    virtual bool              DeleteRange       (
                                                  UIndex     nFirst,
                                                  UIndex     nLast,
                                                  Allocator& rAllocator
                                                );
    virtual void              RemoveLast        (T& rRet, Allocator& rAllocator);
    virtual UIndex            CutTailFromIndex  (
                                                  UIndex     nUIndex,
                                                  Allocator& rAllocator
                                                );
    virtual bool              UniteWith         (
                                                  BTreeArrayBlockInterfaceType* pRight,
                                                  Allocator&                    rAllocator
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              RemoveUpper       ();
  //----------------------------------------------------------------------------
  private:
    inline  T                 RemoveLastItem    (Allocator& rAllocator);
            void              DeleteItem        (
                                                  BIndex     nBIndex,
                                                  T&         rReturn,
                                                  Allocator& rAllocator
                                                );
            T                 RemoveFirstInsertItem
                                                (BIndex nBIndex, const T& rElemt);
            void              AddItem           (
                                                  const T&   rElemt,
                                                  Allocator& rAllocator
                                                );
            void              PlusOneItem       (Allocator& rAllocator);
            void              InsertItem        (
                                                  BIndex     nBIndex,
                                                  const T&   rElemt,
                                                  Allocator& rAllocator
                                                );
    inline  bool              IsFull            () const;
    inline  void              IncreaseUCount    ();
    inline  void              IncreaseArrayCount();
    inline  void              DecreaseArrayCount();
    inline  void              DecreaseUCount    ();
    inline  void              DecreaseUCount    (UIndex nCount);
            void              MoveItems         (
                                                  BIndex nTo,
                                                  BIndex nFrom,
                                                  BIndex nCount
                                                );
            void              DestroyItems      (
                                                  BIndex     nFirst,
                                                  BIndex     nLastPlusOne, 
                                                  Allocator& rAllocator
                                                );
  private:
            BIndex            m_nBCount;
            T*                m_pItems;
            BTreeArrayUpBlockInterfaceType*
                              m_pUpNode;
};

//------------------------------------------------------------------------------
//------------------------------------------------- inline function declarations
//--------------------------------------------------------------- public methods

template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::BTreeArrayBlock(const BTreeArrayBlock& rObj, Allocator& rAllocator)
  : m_nBCount(rObj.m_nBCount),
    m_pItems (NULL),
    m_pUpNode(NULL)
{
  m_pItems = rAllocator.allocate(BLOCK_SIZE);
  T*       pTo   = m_pItems;
  const T* pFrom = rObj.m_pItems;

  for(const T* const pFromEnd = pFrom + rObj.m_nBCount; pFromEnd != pFrom; ++pTo, ++pFrom)
    rAllocator.construct(pTo, *pFrom);
}


// create new BTreeArrayBlock and move the second half of items to it
//
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::BTreeArrayBlock
(
  bool             bSecondHalf, // must be true
  BTreeArrayBlock& rFullBlock,  // must be full
  Allocator&       rAllocator
)
  : m_nBCount(BLOCK_SIZE/2),
    m_pItems (NULL),
    m_pUpNode(NULL)
{
  assert(bSecondHalf);
  assert(rFullBlock.IsFull());
  assert(NULL != rFullBlock.m_pItems);

  m_pItems = rAllocator.allocate(BLOCK_SIZE);

  //     pFrom     rFullBlock      count
  // 
  //               0,1
  // 2 - 2 / 2 = 1   x            2/2 = 1
  //               0,1,2
  // 3 - 3 / 2 = 2     x          3/2 = 1
  //               0,1,2,3
  // 4 - 4 / 2 = 2     x x        4/2 = 2
  //               0,1,2,3,4
  // 5 - 5 / 2 = 3       x x      5/2 = 2
  //               0,1,2,3,4,5
  // 6 - 6 / 2 = 3       x x x    6/2 = 3
  //               0,1,2,3,4,5,6
  // 7 - 7 / 2 = 3         x x x  7/2 = 3
  //
  rFullBlock.m_nBCount = BLOCK_SIZE - BLOCK_SIZE/2;

  T* pFrom = rFullBlock.m_pItems + rFullBlock.m_nBCount;
  T* pTo   = m_pItems;

  for(const T* const pEnd = pTo + m_nBCount; pEnd != pTo; ++pTo, ++pFrom)
  {
    rAllocator.construct(pTo, *pFrom);
    rAllocator.destroy  (pFrom);
  }

  assert(0 < m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator = (const BTreeArrayBlock& rObj)
{
  assert(false);//it is not used
  assert(NULL != m_pItems);
  assert(NULL != m_pUpNode);

  Allocator&   rAllocator = m_pUpNode->GetAllocatorRef();
  const BIndex nMinCount  = m_nBCount < rObj.m_nBCount ? m_nBCount : rObj.m_nBCount;

  T*       pTo   = m_pItems;
  const T* pFrom = rObj.m_pItems;
  for(const T* const pFromEnd = pFrom + nMinCount; pFromEnd != pFrom; *pTo++ = *pFrom++);

  if ( m_nBCount < rObj.m_nBCount )
  {
    for(const T* const pFromEnd = rObj.m_pItems + rObj.m_nBCount; pFromEnd != pFrom; ++pTo, ++pFrom)
      rAllocator.construct(pTo, *pFrom);
  }
  else
  {
    for(const T* const pToEnd = m_pItems + m_nBCount; pToEnd != pTo; ++pTo)
      rAllocator.destroy(pTo);
  }
  m_nBCount = rObj.m_nBCount;
  m_pUpNode = NULL;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::BTreeArrayBlock
(
  Allocator& rAllocator, BTreeArrayUpBlockInterfaceType* pUpNode
)
  : m_nBCount(0      ),
    m_pItems (NULL   ),
    m_pUpNode(pUpNode)
{
  m_pItems = rAllocator.allocate(BLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Deallocate(Allocator& rAllocator)
{
  assert(0 == m_nBCount);
  if ( NULL != m_pItems )
  {
    rAllocator.deallocate(m_pItems, BLOCK_SIZE);
    m_pItems = NULL;
  }
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DestroyAndDeallocate(Allocator& rAllocator)
{
  assert(BLOCK_SIZE >= m_nBCount);
  if ( NULL != m_pItems )  // if exception during constructor was rised
  {
    DestroyItems(0, m_nBCount, rAllocator);
    rAllocator.deallocate(m_pItems, BLOCK_SIZE);
    m_pItems = NULL;
  }
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::~BTreeArrayBlock()
{
  assert(NULL == m_pItems);//see DestroyAndDeallocate
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline UIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetUIndex(BIndex nBIndex)
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_pUpNode);

  return nBIndex + m_pUpNode->GetMyIndex(this);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
typename BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ElementReferenceType
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Insert(const T& rElemt, ElementReferenceType& oReference)
{
  assert(this == oReference.m_uRef.m_pBlock);
  assert(oReference.m_nBIndex < BLOCK_SIZE);
  assert(0 <= oReference.m_nBIndex && oReference.m_nBIndex <= m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  Allocator& rAllocator = m_pUpNode->GetAllocatorRef();

  if ( 0 == oReference.m_nBIndex && !IsFull() )
  {
    // GetPrev() is not very fast therefore we check IsFull() to prevent the
    // call of GetPrev() two times
    ElementReferenceType oPrev = m_pUpNode->GetPrev(this);

    assert(BLOCK_SIZE == oPrev.m_nBIndex ? NULL == oPrev.m_uRef.m_pArray
                                         : NULL != oPrev.m_uRef.m_pBlock);

    if ( BLOCK_SIZE != oPrev.m_nBIndex && !oPrev.m_uRef.m_pBlock->IsFull() )
    {
      oPrev.m_uRef.m_pBlock->AddItem(rElemt, rAllocator);
      oPrev.m_uRef.m_pBlock->IncreaseArrayCount();
      oPrev.m_nBIndex++;
      assert(oPrev.m_nBIndex == oPrev.m_uRef.m_pBlock->m_nBCount - 1);
      return oPrev;
    }
  }

  if ( IsFull() )
  {
    ElementReferenceType oPrev = m_pUpNode->GetPrev(this);

    assert(BLOCK_SIZE == oPrev.m_nBIndex ? NULL == oPrev.m_uRef.m_pArray
                                         : NULL != oPrev.m_uRef.m_pBlock);

    if ( BLOCK_SIZE != oPrev.m_nBIndex && !oPrev.m_uRef.m_pBlock->IsFull() )
    {
      if ( 0 == oReference.m_nBIndex )
      {
        oPrev.m_uRef.m_pBlock->AddItem(rElemt, rAllocator);
        oPrev.m_uRef.m_pBlock->IncreaseArrayCount();
        oPrev.m_nBIndex++;
        assert(oPrev.m_nBIndex == oPrev.m_uRef.m_pBlock->m_nBCount - 1);
      }
      else
      {
        assert(0 < oReference.m_nBIndex);// we know this
        T pMoveObj = RemoveFirstInsertItem(BIndex(oPrev.m_nBIndex - 1), rElemt);
        oPrev.m_uRef.m_pBlock->AddItem(pMoveObj, rAllocator);
        oPrev.m_uRef.m_pBlock->IncreaseArrayCount();
        oPrev = oReference;
        oPrev.m_nBIndex--;
      }

      return oPrev;
    }

    ElementReferenceType oNext = m_pUpNode->GetNext(this);

    if ( BLOCK_SIZE != oNext.m_nBIndex && !oNext.m_uRef.m_pBlock->IsFull() )
    {
      oNext.m_uRef.m_pBlock->InsertItem(0, RemoveLastItem(rAllocator), rAllocator);
      InsertItem(oReference.m_nBIndex, rElemt, rAllocator);
      oNext.m_uRef.m_pBlock->IncreaseArrayCount();
      ElementReferenceType oInserted = oReference;
      oReference.m_nBIndex++;
      if ( BLOCK_SIZE == oReference.m_nBIndex )
        oReference = oNext;

      return oInserted;
    }

    BTreeArrayBlock* pNewBlock = new BTreeArrayBlock(true, *this, m_pUpNode->GetAllocatorRef());

    assert(!IsFull());
    assert(!(pNewBlock->IsFull()));

    ElementReferenceType oInserted = oReference;

    if ( oReference.m_nBIndex <= m_nBCount )
    {
      if ( oReference.m_nBIndex == m_nBCount )
        oReference = ElementReferenceType(pNewBlock, 0);
      else
        oReference.m_nBIndex++;

      InsertItem(oInserted.m_nBIndex, rElemt, rAllocator);
    }
    else
    {
      oInserted.m_nBIndex -= m_nBCount;
      pNewBlock->InsertItem(oInserted.m_nBIndex, rElemt, rAllocator);
      oInserted. m_uRef.m_pBlock = pNewBlock;
      oReference.m_uRef.m_pBlock = pNewBlock;
      oReference.m_nBIndex       = oInserted.m_nBIndex + 1;
    }

    assert(oInserted.m_uRef.m_pBlock == oReference.m_uRef.m_pBlock ? oInserted.m_nBIndex + 1 == oReference.m_nBIndex
                                                                   : oInserted.m_nBIndex + 1 == oInserted.m_uRef.m_pBlock->m_nBCount
                                                                     &&
                                                                     0 == oReference.m_nBIndex
          );
    assert(oReference.m_nBIndex < oReference.m_uRef.m_pBlock->m_nBCount);
    assert(NULL != m_pUpNode);

    m_pUpNode->InsertAfter(this, pNewBlock);
    IncreaseArrayCount();
    return oInserted;
  }

  ElementReferenceType oInserted(oReference);
  InsertItem(oReference.m_nBIndex, rElemt, rAllocator);
  oReference.m_nBIndex++;
  IncreaseArrayCount();
  assert(oInserted.m_nBIndex + 1 == oReference.m_nBIndex);
  assert(oReference.m_nBIndex < m_nBCount);

  return oInserted;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
typename BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ElementReferenceType
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Erase(ElementReferenceType& oReference)
{
  assert(this == oReference.m_uRef.m_pBlock);
  assert(oReference.m_nBIndex < BLOCK_SIZE);
  assert(0 <= oReference.m_nBIndex && oReference.m_nBIndex < m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
  assert(NULL != m_pUpNode);

  T          tElemt;
  Allocator& rAllocator = m_pUpNode->GetAllocatorRef();
  DeleteItem(oReference.m_nBIndex, tElemt, rAllocator);
  DecreaseArrayCount();
  if ( oReference.m_nBIndex == m_nBCount ) 
  {
    // shift the referens to the next element, now it has the
    // index of the deleted element
    oReference = m_pUpNode->GetNext(this);

    if ( 0 == m_nBCount )
    {
      DestroyAndDeallocate(rAllocator);
      m_pUpNode->DeleteItem(this);
    }
  }
  else
    assert(0 < m_nBCount);

  return oReference;
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* 
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Clone(Allocator& rAllocator) const
{
  return new ThisType(*this, rAllocator);
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::SetUpNode(BTreeArrayUpBlockInterfaceType* pUpNode)
{
  assert(BLOCK_SIZE >= m_nBCount);

  m_pUpNode = pUpNode;
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Insert
(
  UIndex                        nUIndex,
  const T&                      rElemt,
  BTreeArrayBlockInterfaceType* pPrev,
  BTreeArrayBlockInterfaceType* pNext,
  Allocator&                    rAllocator
)
{
  assert(nUIndex != UIndex(m_nBCount));// Not Last !
  assert(0 <= nUIndex && nUIndex < UIndex(m_nBCount));
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  BIndex nBIndex = static_cast<BIndex>(nUIndex);
  assert(UIndex(nBIndex) == nUIndex);

  if ( 0 == nBIndex && NULL != pPrev )
  {
    BTreeArrayBlock* pPrevBlock = static_cast<BTreeArrayBlock*>(pPrev);
    if ( !pPrevBlock->IsFull() )
    {
      pPrevBlock->AddItem(rElemt, rAllocator);
      pPrevBlock->IncreaseUCount();
      return NULL;
    }
  }

  if ( IsFull() )
  {
    BTreeArrayBlock* pPrevBlock = static_cast<BTreeArrayBlock*>(pPrev);

    if ( NULL != pPrevBlock && !pPrevBlock->IsFull() )
    {
      assert(0 < nBIndex);// we know this
      T pMoveObj = RemoveFirstInsertItem(BIndex(nBIndex - 1), rElemt);
      pPrevBlock->AddItem(pMoveObj, rAllocator);
      pPrevBlock->IncreaseUCount();

      return NULL;
    }

    BTreeArrayBlock* pNextBlock = static_cast<BTreeArrayBlock*>(pNext);

    if ( NULL != pNext && !pNextBlock->IsFull() )
    {
      pNextBlock->InsertItem(0, RemoveLastItem(rAllocator), rAllocator);
      InsertItem(nBIndex, rElemt, rAllocator);
      pNextBlock->IncreaseUCount();

      return NULL;
    }

    BTreeArrayBlock* pNewBlock = new BTreeArrayBlock(true, *this, rAllocator);

    assert(!IsFull());
    assert(!(pNewBlock->IsFull()));

    if ( nBIndex <= m_nBCount )
      InsertItem(nBIndex, rElemt, rAllocator);
    else
      pNewBlock->InsertItem(BIndex(nBIndex - m_nBCount), rElemt, rAllocator);

    assert(NULL != m_pUpNode);

    return pNewBlock;
  }

  InsertItem(nBIndex, rElemt, rAllocator);
  IncreaseUCount();

  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
T BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Replace(UIndex nUIndex, const T& rBy)
{
  assert(0 < m_nBCount);
  assert(0 <= nUIndex && nUIndex < UIndex(m_nBCount));
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  T tRet = m_pItems[nUIndex];
  m_pItems[nUIndex] = rBy;
  return tRet;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
UIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetCount()
{
  assert(NULL != m_pItems);
  assert(BLOCK_SIZE >= m_nBCount);

  return m_nBCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
BIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetItemsCount() const
{
  return m_nBCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
UIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Depth() const
{
  return 0;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::PushFront(const T& rElemt, Allocator& rAllocator)
{
  assert(BLOCK_SIZE >= m_nBCount);

  if ( IsFull() )
  {
    BTreeArrayBlock* pNewLeaf = new BTreeArrayBlock(rAllocator);//create prev sibling
    pNewLeaf->AddItem(rElemt, rAllocator);

    return pNewLeaf;
  }

  InsertItem(0, rElemt, rAllocator);
  IncreaseUCount();
  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* 
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Add(const T& rElemt, Allocator& rAllocator)
{
  assert(BLOCK_SIZE >= m_nBCount);

  if ( IsFull() )
  {
    BTreeArrayBlock* pNewLeaf = new BTreeArrayBlock(rAllocator);//create next sibling
    pNewLeaf->AddItem(rElemt, rAllocator);

    return pNewLeaf;
  }

  AddItem(rElemt, rAllocator);
  IncreaseUCount();
  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> 
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* 
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::PlusOne(Allocator& rAllocator)
{
  assert(BLOCK_SIZE >= m_nBCount);

  if ( IsFull() )
  {
    BTreeArrayBlock* pNewLeaf = new BTreeArrayBlock(rAllocator);//create next sibling
    pNewLeaf->PlusOneItem(rAllocator);

    return pNewLeaf;
  }

  PlusOneItem(rAllocator);
  IncreaseUCount();
  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
T& BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Get(UIndex nUIndex)
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 <= nUIndex && nUIndex < UIndex(m_nBCount));
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  return m_pItems[nUIndex];
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
T& BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetFirst()
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 < m_nBCount);
  assert(NULL != m_pItems);

  return m_pItems[0];
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
T& BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetLast()
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 < m_nBCount);
  assert(NULL != m_pItems);

  return m_pItems[m_nBCount - 1];
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Begin()
{
  assert(0 < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  return ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, 0);
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Last()
{
  assert(0 < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  return ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, m_nBCount - 1);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Prev
(
  ElementReferenceType& rElement
)
{
  assert(BLOCK_SIZE != rElement.m_nBIndex && "The rElement is past-the-end iterator");

  assert(this == rElement.m_uRef.m_pBlock);
  assert(0 <= rElement.m_nBIndex && rElement.m_nBIndex < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  if ( 0 < rElement.m_nBIndex )
    rElement.m_nBIndex--;
  else
    rElement = m_pUpNode->GetPrev(this);

  return rElement;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Next
(
  ElementReferenceType& rElement
)
{
  assert(BLOCK_SIZE != rElement.m_nBIndex && "The rElement is past-the-end iterator");

  assert(this == rElement.m_uRef.m_pBlock);
  assert(0 <= rElement.m_nBIndex && rElement.m_nBIndex < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  rElement.m_nBIndex++;
  if ( rElement.m_nBIndex < m_nBCount )
    return rElement;

  rElement = m_pUpNode->GetNext(this);
  return rElement;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>& BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Next
(
  ElementReferenceType& rElement, ptrdiff_t nOffset
)
{
  assert(BLOCK_SIZE != rElement.m_nBIndex && "The rElement is past-the-end iterator");

  assert(this == rElement.m_uRef.m_pBlock);
  assert(0 <= rElement.m_nBIndex && rElement.m_nBIndex < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  // 
  //                  -7  -6  -5  -4    -3  -2  -1   0  +1  +2  +3  +4   nOffset
  // +---+---+---+---+---+---+---+---+ +---+---+---+---+---+---+---+---+
  // | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 
  // +---+---+---+---+---+---+---+---+ +---+---+---+---+---+---+---+---+
  //                           ^                     ^
  //                           |                     |
  //                           |                     rElement.m_nBIndex
  //                           |
  //                           rElement.m_nBIndex + nOffset + BLOCK_SIZE
  //                         6 =                3 +    (-5) + 8

  ptrdiff_t nNewIndex = rElement.m_nBIndex + nOffset;
  if ( 0 <= nNewIndex )
  {
    if ( nNewIndex < m_nBCount )
    {
      rElement.m_nBIndex = static_cast<BIndex>(nNewIndex);
      assert(rElement.m_nBIndex == nNewIndex);
      return rElement;
    }
    nNewIndex -= m_nBCount;
  }

  rElement = m_pUpNode->GetNext(this, nNewIndex);
  return rElement;
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetElementReference(UIndex nUIndex)
{
  assert(0 < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 <= nUIndex && nUIndex < static_cast<UIndex>(m_nBCount));

  return ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, static_cast<BIndex>(nUIndex));
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Constate()
{
  assert(0 < m_nBCount && m_nBCount <= BLOCK_SIZE);
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ShrinkToFit(Allocator& rAllocator)
{
  assert(0 < m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  while( BLOCK_SIZE > m_nBCount )
  {
    ElementReferenceType oNext = m_pUpNode->GetNext(this);

    if ( BLOCK_SIZE == oNext.m_nBIndex )
      return;// this block is last one

    assert(NULL != oNext.m_uRef.m_pBlock);
    assert(NULL != oNext.m_uRef.m_pBlock->m_pUpNode);

    T*           pTo      = m_pItems + m_nBCount;
    const T*     pFrom    = oNext.m_uRef.m_pBlock->m_pItems;
    const BIndex nMove    = std::min<BIndex>(BLOCK_SIZE - m_nBCount, oNext.m_uRef.m_pBlock->m_nBCount);
    const T*     pFromEnd = pFrom + nMove;

    assert(0 < nMove);

    while(pFromEnd != pFrom)
    {
      rAllocator.construct(pTo, *pFrom);
      ++pTo;
      ++pFrom;
    }

    m_nBCount += nMove;
    m_pUpNode->UndefineUCount();
    oNext.m_uRef.m_pBlock->m_pUpNode->UndefineUCount();

    if ( nMove < oNext.m_uRef.m_pBlock->m_nBCount )
    {
      pTo      = oNext.m_uRef.m_pBlock->m_pItems;
      pFrom    = pFromEnd;
      pFromEnd = pTo + oNext.m_uRef.m_pBlock->m_nBCount;

      while(pFromEnd != pFrom)
        *pTo++ = *pFrom++;

      assert(pTo == oNext.m_uRef.m_pBlock->m_pItems + oNext.m_uRef.m_pBlock->m_nBCount - nMove);
      assert(pTo + nMove == pFromEnd);
      for(;pFromEnd != pTo; pTo++)
        rAllocator.destroy(pTo);

      oNext.m_uRef.m_pBlock->m_nBCount -= nMove;
    }
    else
    {
      assert(nMove == oNext.m_uRef.m_pBlock->m_nBCount);
      oNext.m_uRef.m_pBlock->DestroyAndDeallocate(rAllocator);
      oNext.m_uRef.m_pBlock->m_nBCount = 0;//to delete pRight
      oNext.m_uRef.m_pBlock->m_pUpNode->DeleteItem(oNext.m_uRef.m_pBlock);
    }
  }
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
UIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Delete
(
  UIndex     nUIndex,
  T&         rTRet,
  bool&      bSubBlockPacked,
  Allocator& rAllocator
)
{
  assert(0 < m_nBCount);
  assert(0 <= nUIndex && nUIndex < UIndex(m_nBCount));
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  const BIndex nBIndex = static_cast<BIndex>(nUIndex);
  assert(UIndex(nBIndex) == nUIndex);
  DecreaseUCount();

  std::swap(m_pItems[nBIndex], rTRet);
  m_nBCount--;

  if ( m_nBCount != nBIndex )
    MoveItems(nBIndex, BIndex(nBIndex + 1), BIndex(m_nBCount - nBIndex));
  rAllocator.destroy(m_pItems + m_nBCount);

  assert(0 <= m_nBCount);
  assert(BLOCK_SIZE > m_nBCount);

  bSubBlockPacked = false;
  return 0;//level
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
bool BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DeleteRange
(
  UIndex nFirst, UIndex nLast, Allocator& rAllocator
)
{
  assert(0 < m_nBCount);
  assert(0 <= nFirst && nFirst <= nLast && nLast < UIndex(m_nBCount));
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  const BIndex nFrom  = static_cast<BIndex>(nFirst);
  const BIndex nUpTo  = static_cast<BIndex>(nLast);
  const BIndex nCount = static_cast<BIndex>(nLast - nFirst + 1);
  assert(UIndex(nFrom) == nFirst);
  assert(UIndex(nUpTo) == nLast);

  if ( nCount == m_nBCount )
  {
    DestroyItems(0, m_nBCount, rAllocator);
    m_nBCount = 0;
    return true;
  }

  const BIndex nMoved = BIndex(m_nBCount - nUpTo - 1);
  MoveItems(nFrom, nUpTo + 1, nMoved);
  DestroyItems(nFrom + nMoved, m_nBCount, rAllocator);

  m_nBCount -= nCount;

  assert(0 < m_nBCount && m_nBCount < BLOCK_SIZE);
  return true;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::RemoveLast(T& rRet, Allocator& rAllocator)
{
  assert(0 < m_nBCount && m_nBCount <= BLOCK_SIZE);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  DecreaseUCount();
  rRet = RemoveLastItem(rAllocator);

  assert(0 <= m_nBCount && m_nBCount < BLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
UIndex BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::CutTailFromIndex(UIndex nUIndex, Allocator& rAllocator)
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 <= nUIndex && nUIndex < UIndex(m_nBCount));
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);
  assert(static_cast<UIndex>(static_cast<BIndex>(nUIndex)) == nUIndex);

  BIndex nBIndex = static_cast<BIndex>(nUIndex);
  DestroyItems(nBIndex, m_nBCount, rAllocator);
  m_nBCount = nBIndex;

  assert(0 <= m_nBCount && m_nBCount < BLOCK_SIZE);

  return 0;//level
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
bool BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::UniteWith
(
  BTreeArrayBlockInterfaceType* pRight,
  Allocator&                    rAllocator
)
{
  assert(NULL != pRight);
  assert(this != pRight);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);
  assert(0 < m_nBCount && m_nBCount <= BLOCK_SIZE);
  assert(0 < pRight->GetItemsCount() && pRight->GetItemsCount() <= BLOCK_SIZE);

  if ( BLOCK_SIZE <= size_t(m_nBCount) )
    return false;

  BIndex                 nMovCount   = BLOCK_SIZE - m_nBCount;
  const BIndex           nRCount     = pRight->GetItemsCount();
  BTreeArrayBlock* const pRightBlock = static_cast<BTreeArrayBlock*>(pRight);

  T* pTo   = m_pItems + m_nBCount;
  T* pFrom = pRightBlock->m_pItems;

  assert(NULL != pFrom);

  if ( nRCount <= nMovCount )
  {
    for(const T* const pFromEnd = pFrom + nRCount; pFromEnd != pFrom; pTo++, pFrom++)
    {
      rAllocator.construct(pTo, *pFrom);
      rAllocator.destroy(pFrom);
    }
    m_nBCount += nRCount;
    pRightBlock->m_nBCount = 0;//to delete pRight
    pRightBlock->DestroyAndDeallocate(rAllocator);

    assert(m_nBCount <= BLOCK_SIZE);
    return true;
  }

  for(const T* const pFromEnd = pFrom + nMovCount; pFromEnd != pFrom; pTo++, pFrom++)
    *pTo = *pFrom;

  pTo = pRightBlock->m_pItems;
  for(const T* const pFromEnd = pTo + nRCount; pFromEnd != pFrom; pTo++, pFrom++)
  {
    *pTo = *pFrom;// can be optimized for simple types
    rAllocator.destroy(pFrom);
  }

  assert(m_nBCount + nMovCount == BLOCK_SIZE);
  assert(0 < nRCount - nMovCount);

  m_nBCount              = BLOCK_SIZE;
  pRightBlock->m_nBCount = nRCount - nMovCount;

  return false;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* 
BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::RemoveUpper()
{
  return this;
}


//-------------------------------------------------------------- private methods


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline T BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::RemoveLastItem(Allocator& rAllocator)
{
  assert(0 != m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  m_nBCount--;

  T tRet(m_pItems[m_nBCount]);
  rAllocator.destroy(m_pItems + m_nBCount);
  return tRet;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DeleteItem(BIndex nBIndex, T& rElemt, Allocator& rAllocator)
{
  std::swap(m_pItems[nBIndex], rElemt);
  m_nBCount--;

  if ( m_nBCount != nBIndex )
    MoveItems(nBIndex, BIndex(nBIndex + 1), BIndex(m_nBCount - nBIndex));
  rAllocator.destroy(m_pItems + m_nBCount);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
T BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::RemoveFirstInsertItem(BIndex nBIndex, const T& rElemt)
{
  assert(0 <= nBIndex);
  assert(nBIndex < m_nBCount);
  assert(0 != m_nBCount);
  assert(BLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);
  assert(NULL != m_pItems);

  T tRet = m_pItems[0];

  MoveItems(0, 1, nBIndex);
  m_pItems[nBIndex] = rElemt;

  return tRet;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::AddItem(const T& rElemt, Allocator& rAllocator)
{
  assert(BLOCK_SIZE > m_nBCount);
  assert(!IsFull());
  assert(NULL != m_pItems);

  rAllocator.construct(m_pItems + m_nBCount, rElemt);
  m_nBCount++;

  assert(0 < m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::PlusOneItem(Allocator& rAllocator)
{
  assert(BLOCK_SIZE > m_nBCount);
  assert(!IsFull());

  rAllocator.construct(m_pItems + m_nBCount, T());
  m_nBCount++;

  assert(0 < m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::InsertItem
(
  BIndex     nBIndex,
  const T&   rElemt,
  Allocator& rAllocator
)
{
  assert(!IsFull());
  assert(0 <= nBIndex);
  assert(m_nBCount >= nBIndex);//bad index?
  assert(BLOCK_SIZE > m_nBCount);

  if ( nBIndex != m_nBCount )
  {
    assert(0 < m_nBCount);
    rAllocator.construct(m_pItems + m_nBCount, m_pItems[m_nBCount - 1]);
    MoveItems(BIndex(nBIndex + 1), nBIndex, BIndex(m_nBCount - nBIndex - 1));
    m_pItems[nBIndex] = rElemt;
  }
  else
    rAllocator.construct(m_pItems + m_nBCount, rElemt);

  m_nBCount++;

  assert(0 < m_nBCount);
  assert(m_nBCount <= BLOCK_SIZE);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline bool BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::IsFull() const
{
  assert(BLOCK_SIZE >= m_nBCount);
  assert(0 <= m_nBCount);

  return BLOCK_SIZE == m_nBCount;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::IncreaseUCount()
{
  assert(NULL != m_pUpNode);

  m_pUpNode->IncreaseUCount();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::IncreaseArrayCount()
{
  assert(NULL != m_pUpNode);

  m_pUpNode->IncreaseArrayCount();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DecreaseArrayCount()
{
  assert(NULL != m_pUpNode);

  m_pUpNode->DecreaseArrayCount();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DecreaseUCount()
{
  assert(NULL != m_pUpNode);

  m_pUpNode->DecreaseUCount();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DecreaseUCount(UIndex nCount)
{
  assert(NULL != m_pUpNode);
  assert(0 < nCount && nCount <= BLOCK_SIZE);

  m_pUpNode->DecreaseUCount(nCount);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::MoveItems(BIndex nTo, BIndex nFrom, BIndex nCount)
{
  if ( nCount <= 0 )
    return;

  assert(0 <= nTo);
  assert(nTo < BLOCK_SIZE);
  assert(nTo + nCount <= BLOCK_SIZE);
  assert(0 <= nFrom);
  assert(nFrom < BLOCK_SIZE);
  assert(nFrom + nCount <= BLOCK_SIZE);
  assert(nCount <= m_nBCount);

  if ( nTo > nFrom )
  {
    //                 7         
    //          |<--nCount--->|     
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  | | | | |1|2|3|4|5|6|7| | | | | | | | | | | | | |   before
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  | | | | | | | | | | | | | | |1|2|3|4|5|6|7| | | |   after
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //           ^           ^       ^             ^
    //           |           |       |<--nCount--->|
    //           nFrom       |       nTo           nTo+nCount
    //                       |
    //                       nFrom + nCount - 1
    //  
    //   or
    //  
    //                       nFrom + nCount - 1
    //                       |
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  | | | | |1|2|3|4|5|6|7| | | | | | | | | | | | | |   before
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  | | | | | |1|2|3|4|5|6|7| | | | | | | | | | | | |   after
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //           ^ ^             ^
    //           | |<--nCount--->|
    //           | |             |
    //           | nTo           nTo+nCount
    //           |
    //           nFrom       
    //                     
    //
    T*             pTo   = &m_pItems[nTo   + nCount - 1];
    T*             pFrom = &m_pItems[nFrom + nCount - 1];
    const T* const pStop = &m_pItems[nTo - 1];

    while(pStop != pTo)
      *pTo-- = *pFrom--;

    return;
  }

  if ( nTo < nFrom )
  {
    T* pTo   = &m_pItems[nTo];
    T* pFrom = &m_pItems[nFrom];
    for(T* const pLimit = pFrom + nCount; pLimit != pFrom; *pTo++ = *pFrom++);
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
void BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DestroyItems
(
  BIndex     nFirst,
  BIndex     nLastPlusOne, 
  Allocator& rAllocator
)
{
  assert((0 <= nFirst && nFirst < nLastPlusOne) || (0 == nFirst && 0 == nLastPlusOne));
  assert(NULL != m_pItems);

  T* pIt = m_pItems + nFirst;
  for(const T* const pEnd = m_pItems + nLastPlusOne; pEnd != pIt; ++pIt)
    rAllocator.destroy(pIt);
}


//------------------------------------------------------------- inline functions
}// namespace Detail
}// namespace FastCpp
#endif//FastCpp_BTREEARRAYBLOCK_HPP
//** (END OF FILE  : BTreeArrayBlock.hpp) **************************************
