//******************************************************************************
// FILE..........: BTreeArrayUpBlock.hpp
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- -------------------------------------------------
//                     y  m  d
//                 2012-11-19 Created by VDima.
//                 2013-06-28 Renamed BTreeArrayDetail.hpp -> BTreeArrayUpBlock.hpp
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_BTREEARRAYUPBLOCK_HPP
#define FastCpp_BTREEARRAYUPBLOCK_HPP

//--------------------------------------------------------------- import std lib
#include <assert.h>
//-----------------------------------------------------------------------------
#ifndef                     FastCpp_ELEMENTREFERENCE_HPP
#include <FastCpp/BTreeArray/detail/ElementReference.hpp>
#endif//                    FastCpp_ELEMENTREFERENCE_HPP


#ifndef                     FastCpp_BTREEARRAYBLOCK_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayBlock.hpp>
#endif//                    FastCpp_BTREEARRAYBLOCK_HPP

#ifndef                     FastCpp_BTREEARRAYINTERFACE_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayInterface.hpp>
#endif//                    FastCpp_BTREEARRAYINTERFACE_HPP

//-----------------------------------------------------------------------------
//| |       |                 |                 |                            |
namespace FastCpp
{
namespace Detail
{


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
class BTreeArrayUpBlock : public BTreeArrayBlockInterface  <T, UIndex, BIndex, BLOCK_SIZE, Allocator>,
                          public BTreeArrayUpBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
{
  public:
    typedef BTreeArrayBlockInterface  <T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayBlockInterfaceType;
    typedef BTreeArrayUpBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayUpBlockInterfaceType;
    typedef ElementReference          <T, UIndex, BIndex, BLOCK_SIZE, Allocator> ElementReferenceType;
    typedef BTreeArrayUpBlock                                                    ThisType;
  private:
    inline                    BTreeArrayUpBlock (const BTreeArrayUpBlock&);// = deleted
    inline                    BTreeArrayUpBlock (const BTreeArrayUpBlock& rObj, Allocator& rAllocator);
    inline  void              operator =        (const BTreeArrayUpBlock& rObj);

  public:
    inline                    BTreeArrayUpBlock ();
    inline                    BTreeArrayUpBlock (
                                                  BTreeArrayUpBlockInterfaceType* p,
                                                  UIndex                          nUCount
                                                );
    virtual                  ~BTreeArrayUpBlock ();

  //----------------------------------------------------BTreeArrayBlockInterface
  public:
    virtual BTreeArrayBlockInterfaceType*
                              Clone             (Allocator& rAllocator) const;
    virtual void              Deallocate        (Allocator& rAllocator);
    virtual void              DestroyAndDeallocate
                                                (Allocator& rAllocator);
    virtual void              SetUpNode         (BTreeArrayUpBlockInterfaceType* pUNode);
    virtual BTreeArrayBlockInterfaceType*
                              Insert            (
                                                  UIndex                        nUIndex,
                                                  const T&                      rElemt,
                                                  BTreeArrayBlockInterfaceType* pArrPrev,
                                                  BTreeArrayBlockInterfaceType* pNext,
                                                  Allocator&                    rAllocator
                                                );

    virtual T                 Replace           (UIndex Index, const T& rBy);

    virtual UIndex            GetCount          ();
    virtual BIndex            GetItemsCount     () const;
    virtual UIndex            Depth             () const;
    virtual BTreeArrayBlockInterfaceType*
                              PushFront         (const T& rElemt, Allocator& rAllocator);
    virtual BTreeArrayBlockInterfaceType*
                              Add               (const T& rElemt, Allocator& rAllocator);
    virtual BTreeArrayBlockInterfaceType*
                              PlusOne           (Allocator& rAllocator);
    virtual T&                Get               (UIndex Index);
    virtual T&                GetFirst          ();
    virtual T&                GetLast           ();

    virtual ElementReferenceType
                              Begin             ();
    virtual ElementReferenceType
                              Last              ();
    virtual ElementReferenceType
                              GetElementReference
                                                (UIndex nUIndex);
    virtual BTreeArrayBlockInterfaceType*
                              GetFirstBlockOnLevel
                                                (UIndex nULevel);
    virtual BTreeArrayBlockInterfaceType*
                              GetLastBlockOnLevel
                                                (UIndex nULevel);
    virtual void              Constate          ();
    virtual void              ShrinkToFit       (Allocator& rAllocator);
    virtual UIndex            Delete            (
                                                  UIndex     nUIndex,
                                                  T&         rRet,
                                                  bool&      bSubBlockPacked,
                                                  Allocator& rAllocator
                                                );
    virtual bool              DeleteRange       (
                                                  UIndex     nFirst,
                                                  UIndex     nLast,
                                                  Allocator& rAllocator
                                                );
    virtual void              RemoveLast        (T& rRet, Allocator& rAllocator);
    virtual UIndex            CutTailFromIndex  (
                                                  UIndex     nUIndex,
                                                  Allocator& rAllocator
                                                );
    virtual bool              UniteWith         (
                                                  BTreeArrayBlockInterfaceType* pRight,
                                                  Allocator&                    rAllocator
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              RemoveUpper       ();
  //--------------------------------------------------BTreeArrayUpBlockInterface
  public:
    virtual void              IncreaseUCount    ();
    virtual void              IncreaseArrayCount();
    virtual void              DecreaseArrayCount();
    virtual void              DecreaseUCount    ();
    virtual void              DecreaseUCount    (UIndex nCount);
    virtual void              UndefineUCount    ();
    virtual UIndex            GetMyIndex        (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetPrev           (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetNext           (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetNext           (
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  ptrdiff_t                     nOffset
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              GetPrevSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              GetNextSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                );

    virtual void              DeleteItem        (BTreeArrayBlockInterfaceType* pBlock);
    virtual void              InsertAfter       (
                                                  BTreeArrayBlockInterfaceType* pAfterThis,
                                                  BTreeArrayBlockInterfaceType* pNewBlock
                                                );
    virtual Allocator&        GetAllocatorRef   ();
  //----------------------------------------------------------------------------
  public:
            void              RecalcIndexes     (
                                                  UIndex& nUIndex,
                                                  BIndex& nBIndex
                                                );
            void              GetHalveFrom      (BTreeArrayUpBlock& rFullBlock);
            void              Fusion            (Allocator&         rAllocator);
    inline  bool              IsFull            () const;
    inline  BTreeArrayBlockInterfaceType*
                              RemoveLastItem    ();
            void              InsertItem        (
                                                  BIndex                        nBIndex,
                                                  BTreeArrayBlockInterfaceType* pObj
                                                );
            void              DeleteItem        (BIndex nBIndex);
            void              AddItem           (BTreeArrayBlockInterfaceType* pObj);
  private:
            bool              PackSubblocks     (Allocator& rAllocator);
            void              Pack              (
                                                  UIndex     nULevel,
                                                  bool&      bSubBlockPacked,
                                                  Allocator& rAllocator
                                                );
            BTreeArrayBlockInterfaceType* 
                              RemoveFirstInsertItem
                                                (
                                                  BIndex                        nBIndex,
                                                  BTreeArrayBlockInterfaceType* pItem
                                                );

  private:
            BIndex            m_nBCount;
            UIndex            m_nUCount;

            // In the next line the "error C2466: cannot allocate an array of
            // constant size 0" or the "error C2118: negative subscript" means,
            // that the UPBLOCK_SIZE value was truncated  
            // for the reason of specialization: the specified value
            // of UPBLOCK_SIZE is bigger of the maximal value of BIndex type
            // Example:
            //   FastCpp::BTreeArray<int, unsigned int, signed char, 8, 214> arrT;
            // here
            //   (signed char) 214 == -42
            //
            BTreeArrayBlockInterfaceType*
                              m_arrItems[UPBLOCK_SIZE];
            BTreeArrayUpBlockInterfaceType*
                              m_pUpNode;
};


//----------------------------------------------BTreeArrayBlockInterface methods


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::BTreeArrayUpBlock
(
  const BTreeArrayUpBlock& rObj,
 Allocator&                rAllocator
)
  : m_nBCount(rObj.m_nBCount),
    m_nUCount(rObj.m_nUCount),
    m_pUpNode(NULL          )
{
  const BTreeArrayBlockInterfaceType* const* pFrom = rObj.m_arrItems;
  BTreeArrayBlockInterfaceType**             pTo   = m_arrItems;

  for(
      const BTreeArrayBlockInterfaceType* const* const pEnd = pFrom + rObj.m_nBCount;
      pEnd != pFrom;
      pFrom++, pTo++
     )
  {
    assert(NULL != pFrom);
    assert(NULL != pTo);
    assert(NULL != *pFrom);
    *pTo = (*pFrom)->Clone(rAllocator);

    assert(NULL != *pTo);
    (*pTo)->SetUpNode(this);
  }
}


//template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
//inline void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::operator = (const BTreeArrayUpBlock&)
//{
//  assert(false);
//}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::BTreeArrayUpBlock()
  : m_nBCount(0),
    m_nUCount(UINDEX_LIMIT),
    m_pUpNode(NULL)
{
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::BTreeArrayUpBlock
(
  BTreeArrayUpBlockInterfaceType* pUpNode,
  UIndex                          nUCount
)
  : m_nBCount(0      ),
    m_nUCount(nUCount),
    m_pUpNode(pUpNode)
{
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::~BTreeArrayUpBlock()
{
  assert(0 <= m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);

  while( 0 != m_nBCount )
    delete m_arrItems[--m_nBCount];
}


//nBIndex - index of subcollection
//nUIndex - index of object in subcollection 
//
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::RecalcIndexes(UIndex& nUIndex, BIndex& rBIndex)
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);

  if ( UINDEX_LIMIT == m_nUCount || (m_nUCount / 2) >= nUIndex )
  {
    UIndex uCount = 0;
    UIndex uItmCount;
    
    for(rBIndex = 0; rBIndex < m_nBCount - 1; rBIndex++)
    {
      uItmCount = m_arrItems[rBIndex]->GetCount();
      if ( nUIndex < uCount + uItmCount )
      {
        nUIndex -= uCount;
        return;
      }
      uCount += uItmCount;
    }

    m_nUCount = uCount + m_arrItems[rBIndex]->GetCount();

    assert(nUIndex < m_nUCount);

    nUIndex -= uCount;
    return;
  }

  assert(nUIndex < m_nUCount);

  UIndex uCount = m_nUCount;
  rBIndex = m_nBCount;

  assert(0 < rBIndex);

  while(0 != rBIndex)
  {
    uCount -= m_arrItems[--rBIndex]->GetCount();

    if ( nUIndex >= uCount )
    {
      nUIndex -= uCount;
      return;
    }
  }
  assert(false);//bad index?
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetHalveFrom(BTreeArrayUpBlock& rFullBlock)
{
  assert(this != &rFullBlock);
  assert(rFullBlock.IsFull());
  assert(0 == m_nBCount);

  m_nBCount            = UPBLOCK_SIZE/2;
  rFullBlock.m_nBCount = UPBLOCK_SIZE - UPBLOCK_SIZE/2;

  BTreeArrayBlockInterfaceType** pFrom = &rFullBlock.m_arrItems[rFullBlock.m_nBCount];
  BTreeArrayBlockInterfaceType** pTo   = m_arrItems;

  for(const BTreeArrayBlockInterfaceType* const* const pEnd = pTo + m_nBCount; pEnd != pTo; pTo++)
  {
    *pTo = *pFrom++;
    assert(NULL != *pTo);
    (*pTo)->SetUpNode(this);
  }

  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE/2 + 1);
  assert(0 < rFullBlock.m_nBCount && rFullBlock.m_nBCount <= UPBLOCK_SIZE/2 + 1);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Fusion(Allocator& rAllocator)
{
  assert(1 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);

  BIndex nLeft  = 0;

  for(BIndex nRight = 1; nRight < m_nBCount; nRight++)
  {
    if ( m_arrItems[nLeft]->UniteWith(m_arrItems[nRight], rAllocator) )
    {
      assert(0 == m_arrItems[nRight]->GetItemsCount());
    }
    else
      nLeft = nRight;
  }

  nLeft = 0;
  for(BIndex nRight = 1; nRight < m_nBCount; nRight++)
  {
    if ( 0 == m_arrItems[nRight]->GetItemsCount() )
      delete m_arrItems[nRight];
    else
    {
      nLeft++;
      m_arrItems[nLeft] = m_arrItems[nRight];
    }
  }

  m_nBCount = nLeft + 1;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::InsertItem(BIndex nBIndex, BTreeArrayBlockInterfaceType* pObj)
{
  assert(NULL != pObj);
  assert(!IsFull());
  assert(0 <= nBIndex);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(m_nBCount >= nBIndex);//bad index?

  if ( nBIndex != m_nBCount )
  {
    BTreeArrayBlockInterfaceType** pTo   = m_arrItems + m_nBCount;
    BTreeArrayBlockInterfaceType** pFrom = pTo - 1;
    BTreeArrayBlockInterfaceType** pEnd  = m_arrItems + nBIndex;

    while(pEnd != pTo)
      *pTo-- = *pFrom--;
  }

  m_arrItems[nBIndex] = pObj;
  m_nBCount++;
  pObj->SetUpNode(this);

  assert(m_nBCount <= UPBLOCK_SIZE);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DeleteItem(BIndex nBIndex)
{
  assert(0 < m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_arrItems[nBIndex]);
  assert(0 == m_arrItems[nBIndex]->GetItemsCount());
  assert(NULL != m_pUpNode);

  delete m_arrItems[nBIndex];

  m_nBCount--;

  if ( m_nBCount != nBIndex )
  {
    assert(0 < m_nBCount);
    BTreeArrayBlockInterfaceType** pTo   = &m_arrItems[nBIndex];
    BTreeArrayBlockInterfaceType** pFrom = pTo + 1;
    BTreeArrayBlockInterfaceType** pEnd  = m_arrItems + m_nBCount;

    while(pEnd != pTo)
      *pTo++ = *pFrom++;
  }

  assert(0 <= m_nBCount);
  assert(UPBLOCK_SIZE > m_nBCount);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::AddItem(BTreeArrayBlockInterfaceType* pObj)
{
  assert(NULL != pObj);
  assert(!IsFull());
  assert(UPBLOCK_SIZE > m_nBCount);

  m_arrItems[m_nBCount] = pObj;
  m_nBCount++;
  pObj->SetUpNode(this);

  assert(m_nBCount <= UPBLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Clone(Allocator& rAllocator) const
{
  return new ThisType(*this, rAllocator);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Deallocate(Allocator&)// rAllocator
{
  assert(NULL != m_pUpNode);
  assert(0 == m_nBCount);
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DestroyAndDeallocate(Allocator& rAllocator)
{
  assert(NULL != m_pUpNode);
  BTreeArrayBlockInterfaceType** pBlock = m_arrItems;
  for(
      const BTreeArrayBlockInterfaceType *const *const pEnd = m_arrItems + m_nBCount;
      pEnd != pBlock;
      pBlock++
     )
  {
    assert(NULL != pBlock);
    (*pBlock)->DestroyAndDeallocate(rAllocator);
  }
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::SetUpNode(BTreeArrayUpBlockInterfaceType* pUpNode)
{
  assert(UPBLOCK_SIZE >= m_nBCount);
  m_pUpNode = pUpNode;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Insert
(
  UIndex                        nUIndex,
  const T&                      rElemt,
  BTreeArrayBlockInterfaceType* pArrPrev,
  BTreeArrayBlockInterfaceType* pArrNext,
  Allocator&                    rAllocator
)
{
  assert(NULL != pArrNext ? NULL != static_cast<BTreeArrayUpBlock*>(pArrNext)->m_arrItems[0]
                          : true
        );
  assert(NULL != pArrNext ? 0 < static_cast<BTreeArrayUpBlock*>(pArrNext)->m_nBCount : true);
  assert(NULL != pArrPrev ? 0 < static_cast<BTreeArrayUpBlock*>(pArrPrev)->m_nBCount : true);
  assert(NULL != m_pUpNode);

  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_arrItems[nBIndex]);

  // our siblings are the same type
  BTreeArrayUpBlock* pNext = static_cast<BTreeArrayUpBlock*>(pArrNext);
  BTreeArrayUpBlock* pPrev = static_cast<BTreeArrayUpBlock*>(pArrPrev);

  BTreeArrayBlockInterfaceType* pNew = m_arrItems[nBIndex]->Insert
    (
      nUIndex,
      rElemt,
      0 < nBIndex ? m_arrItems[nBIndex - 1]
                  : (NULL != pPrev ? pPrev->m_arrItems[pPrev->m_nBCount - 1] : NULL),

      nBIndex + 1 < m_nBCount ? m_arrItems[nBIndex + 1]
                              : (NULL != pNext ? pNext->m_arrItems[0] : NULL),
      rAllocator
    );

  if ( NULL == pNew )
    return NULL;

  if ( !IsFull() )
  {
    InsertItem(nBIndex + 1, pNew);                  
    IncreaseUCount();

    return NULL;
  }

  if ( NULL != pPrev && !pPrev->IsFull() )
  {
    UndefineUCount();//children count changed

    // nBIndex (not nBIndex+1) because we move m_arrItems[0] to pPrev
    BTreeArrayBlockInterfaceType* pMoveNode = RemoveFirstInsertItem(nBIndex, pNew);
    pPrev->AddItem(pMoveNode);
    pPrev->UndefineUCount();
    return NULL;
  }

  assert(IsFull());

  if ( NULL != pNext && !pNext->IsFull() )
  {
    UndefineUCount();//children count changed

    if ( UPBLOCK_SIZE != nBIndex + 1 )
    {
      BTreeArrayBlockInterfaceType* pMoveNode = RemoveLastItem();
      InsertItem(nBIndex + 1, pNew);
      assert(NULL != pMoveNode);
      pNew = pMoveNode;
    }

    pNext->InsertItem(0, pNew);
    pNext->UndefineUCount();//count changed
    return NULL;
  }

  m_nUCount = UINDEX_LIMIT;

  BTreeArrayUpBlock* pNewColl = new BTreeArrayUpBlock;
  pNewColl->GetHalveFrom(*this);

  assert(!IsFull());
  assert(!(pNewColl->IsFull()));

  if ( nBIndex < m_nBCount )
    InsertItem(nBIndex + 1, pNew);
  else
    pNewColl->InsertItem(BIndex(nBIndex + 1 - m_nBCount), pNew);

  assert(UINDEX_LIMIT == m_nUCount);          //child count changed
  assert(UINDEX_LIMIT == pNewColl->m_nUCount);//child count changed

  return pNewColl;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
T BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Replace(UIndex nUIndex, const T& rBy)
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);

  return m_arrItems[nBIndex]->Replace(nUIndex, rBy);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::PushFront(const T& rElemt, Allocator& rAllocator)
{
  assert(0 < m_nBCount);//empty array? - it is impossible
  assert(m_nBCount <= UPBLOCK_SIZE);
  assert(NULL != m_arrItems[0]);

  BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pNew = m_arrItems[0]->PushFront(rElemt, rAllocator);

  if ( NULL != pNew )
  {
    if ( IsFull() )
    {
      BTreeArrayUpBlock* pPrev = new BTreeArrayUpBlock;

      pPrev->AddItem(pNew);
      pPrev->m_nUCount = 1;

      return pPrev;
    }

    InsertItem(0, pNew);                  
    IncreaseUCount();//We know that pNew block has only one T-element
  }

  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Add(const T& rElemt, Allocator& rAllocator)
{
  assert(0 < m_nBCount);//empty array? - it is impossible
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_arrItems[m_nBCount - 1]);//internal error

  BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pNew = m_arrItems[m_nBCount - 1]->Add(rElemt, rAllocator);

  if ( NULL != pNew )
  {
    if ( IsFull() )
    {
      BTreeArrayUpBlock* pNext = new BTreeArrayUpBlock;

      pNext->AddItem(pNew);
      pNext->m_nUCount = 1;

      return pNext;
    }

    AddItem(pNew);
    IncreaseUCount();//We know that pNew block has only one T-element
  }

  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::PlusOne(Allocator& rAllocator)
{
  assert(0 < m_nBCount);//empty array? - it is impossible - internal error
  assert(UPBLOCK_SIZE >= m_nBCount);//internal error
  assert(NULL != m_arrItems[m_nBCount - 1]);//internal error

  BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pNew = m_arrItems[m_nBCount - 1]->PlusOne(rAllocator);

  if ( NULL != pNew )
  {
    if ( IsFull() )
    {
      BTreeArrayUpBlock* pNext = new BTreeArrayUpBlock;

      pNext->AddItem(pNew);
      pNext->m_nUCount = 1;

      return pNext;
    }

    AddItem(pNew);
    IncreaseUCount();//We know that pNew block has only one T-element
  }

  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
T& BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Get(UIndex nUIndex)
{
  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_pUpNode);

  return m_arrItems[nBIndex]->Get(nUIndex);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
T& BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetFirst()
{
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 < m_nBCount);
  assert(NULL != m_arrItems[0]);

  return m_arrItems[0]->GetFirst();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
T& BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetLast()
{
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 < m_nBCount);
  assert(NULL != m_arrItems[m_nBCount - 1]);

  return m_arrItems[m_nBCount - 1]->GetLast();
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Begin()
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_arrItems[0]);

  return m_arrItems[0]->Begin();
}


//virtual 
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Last()
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_arrItems[m_nBCount - 1]);

  return m_arrItems[m_nBCount - 1]->Last();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetElementReference
(
  UIndex nUIndex
)
{
  assert(0 <= nUIndex);

  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_pUpNode);

  return m_arrItems[nBIndex]->GetElementReference(nUIndex);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetFirstBlockOnLevel
(
  UIndex nULevel
)
{
  assert(0 <= nULevel);
  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE);

  if ( 0 == nULevel )
    return this;

  return m_arrItems[0]->GetFirstBlockOnLevel(nULevel - 1);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetLastBlockOnLevel
(
  UIndex nULevel
)
{
  assert(0 <= nULevel);
  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE);

  if ( 0 == nULevel )
    return this;

  return m_arrItems[m_nBCount - 1]->GetLastBlockOnLevel(nULevel - 1);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Constate()
{
  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE);
  assert(NULL != m_pUpNode);

  for(BIndex i = 0; i < m_nBCount; i++)// optimize this
  {
    assert(NULL != m_arrItems[i]);
    m_arrItems[i]->Constate();
  }
  GetCount();//recalc m_nUCount
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::ShrinkToFit(Allocator& rAllocator)
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_pUpNode);

  for(BIndex i = 0; i < m_nBCount; ++i)
    m_arrItems[i]->ShrinkToFit(rAllocator);//the call can change this->m_nBCount

  BTreeArrayUpBlock* pNext;
  BIndex             nOldBCount;
  while(m_nBCount < UPBLOCK_SIZE)
  {
    pNext = static_cast<BTreeArrayUpBlock*>(m_pUpNode->GetNextSibling(this, 0));
    if ( NULL == pNext )
      return; // This block is last one
    assert(NULL != pNext->m_pUpNode);

    // move siblings
    BTreeArrayBlockInterfaceType** pTo      = m_arrItems + m_nBCount;
    BTreeArrayBlockInterfaceType** pFrom    = pNext->m_arrItems;
    const BIndex                   nMove    = std::min<BIndex>(UPBLOCK_SIZE - m_nBCount, pNext->m_nBCount);
    BTreeArrayBlockInterfaceType** pFromEnd = pFrom + nMove;

    assert(0 < nMove);
    while(pFromEnd != pFrom)
    {
      *pTo = *pFrom;
      (*pTo)->SetUpNode(this);
      ++pTo;
      ++pFrom;
    }
    UndefineUCount();
    pNext->UndefineUCount();
    nOldBCount = m_nBCount;
    m_nBCount += nMove;

    // shift items at the next block
    if ( nMove < pNext->m_nBCount )
    {
      assert(UPBLOCK_SIZE == m_nBCount);

      pTo      = pNext->m_arrItems;
      pFrom    = pFromEnd;
      pFromEnd = pTo + pNext->m_nBCount;

      while(pFromEnd != pFrom)
        *pTo++ = *pFrom++;

      pNext->m_nBCount -= nMove;
    }
    else
    {
      assert(nMove == pNext->m_nBCount);
      pNext->m_nBCount = 0;
      pNext->m_pUpNode->DeleteItem(pNext);
      // pNext is not valid after call 
    }

    // Shrink the moved items
    for(BIndex i = nOldBCount; i < m_nBCount; ++i)
      m_arrItems[i]->ShrinkToFit(rAllocator);//the call can change this->m_nBCount
  }
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
UIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Delete
(
  UIndex     nUIndex,
  T&         rRet,
  bool&      bSubBlockPacked,
  Allocator& rAllocator
)
{
  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_arrItems[nBIndex]);

  BTreeArrayBlockInterfaceType* pChildArr = m_arrItems[nBIndex];
  UIndex                        nLevel    = pChildArr->Delete(nUIndex, rRet, bSubBlockPacked, rAllocator) + 1;

  if ( 0 == pChildArr->GetItemsCount() )
  {
    pChildArr->DestroyAndDeallocate(rAllocator);
    DeleteItem(nBIndex);//pChildArr - not valid after the call
    bSubBlockPacked = true;// upper should pack subblocks, but here 
                           // we should not pack subblocks, because
                           // we remove only one element and whole subblock,
                           // so the capacity of subblocks decreased and the
                           // density of elements increased  
  }
  else 
    Pack(nLevel, bSubBlockPacked, rAllocator);

  assert(NULL != m_pUpNode);
  return nLevel;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
bool BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DeleteRange
(
  UIndex nFirst, UIndex nLast, Allocator& rAllocator
)
{
  assert(nFirst <= nLast);

  const UIndex nNewUCount = UINDEX_LIMIT == m_nUCount ? UINDEX_LIMIT
                                                      : m_nUCount - (nLast - nFirst + 1);
  BIndex nBIndexFirst;
  BIndex nBIndexLast;
  RecalcIndexes(nFirst, nBIndexFirst);
  RecalcIndexes(nLast,  nBIndexLast);

  assert(0 <= nBIndexFirst && nBIndexFirst < m_nBCount);
  assert(NULL != m_arrItems[nBIndexFirst]);
  assert(0 <= nBIndexLast && nBIndexLast < m_nBCount);
  assert(NULL != m_arrItems[nBIndexLast]);

  if ( nBIndexFirst == nBIndexLast )
  {
    BTreeArrayBlockInterfaceType* pBlock = m_arrItems[nBIndexFirst];
    bool bSubBlockPacked = pBlock->DeleteRange(nFirst, nLast, rAllocator);
    m_nUCount = nNewUCount;

    if ( 0 == pBlock->GetItemsCount() )
    {
      pBlock->DestroyAndDeallocate(rAllocator);
      DeleteItem(nBIndexFirst);
      if ( 2 <= m_nBCount )
        PackSubblocks(rAllocator);// ignore return value,
      bSubBlockPacked = true;// upper should pack subblocks anyway
    }
    else
    {
      if ( bSubBlockPacked && 2 <= m_nBCount )
        bSubBlockPacked = PackSubblocks(rAllocator);
    }

    return bSubBlockPacked;
  }

  BIndex nBDelFrom = nBIndexFirst;
  BIndex nBDelLast = nBIndexLast;

  if ( nLast + 1 != m_arrItems[nBIndexLast]->GetCount() )
  {
    m_arrItems[nBIndexLast]->DeleteRange(0, nLast, rAllocator);
    assert(0 < m_arrItems[nBIndexLast]->GetCount());
    --nBDelLast;
  }

  if ( 0 < nFirst )
  {
    m_arrItems[nBIndexFirst]->CutTailFromIndex(nFirst, rAllocator);
    assert(0 < m_arrItems[nBIndexFirst]->GetItemsCount());
    ++nBDelFrom;
  }
  else
    assert(0 == nBIndexFirst);

  assert(nBDelFrom < m_nBCount);
  assert(nBDelLast < m_nBCount);

  if ( nBDelFrom <= nBDelLast )
  {
    for(BIndex i = nBDelFrom; i <= nBDelLast; ++i)
    {
      assert(NULL != m_arrItems[i]);
      m_arrItems[i]->DestroyAndDeallocate(rAllocator);
      delete m_arrItems[i];
    }

    if ( nBDelLast + 1 == m_nBCount )
      m_nBCount = nBDelFrom;
    else
    {
      assert(nBDelFrom < nBDelLast + 1 && nBDelLast + 1 < m_nBCount);
      BTreeArrayBlockInterfaceType** pTo   = &m_arrItems[nBDelFrom];
      BTreeArrayBlockInterfaceType** pFrom = &m_arrItems[nBDelLast + 1];
      BTreeArrayBlockInterfaceType** pEnd  = m_arrItems + m_nBCount;
      while(pEnd != pFrom)
        *pTo++ = *pFrom++;
      m_nBCount -= nBDelLast - nBDelFrom + 1;
    }
  }

  m_nUCount = nNewUCount;
  if ( 2 <= m_nBCount )
    PackSubblocks(rAllocator);
  // upper should pack subblocks anyway
  return true; 
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::RemoveLast(T& rRet, Allocator& rAllocator)
{
  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE);
  assert(NULL != m_arrItems[m_nBCount - 1]);

  BTreeArrayBlockInterfaceType* pChildArr = m_arrItems[m_nBCount - 1];
  pChildArr->RemoveLast(rRet, rAllocator);

  if ( 0 == pChildArr->GetItemsCount() )
  {
    m_nBCount--;
    pChildArr->DestroyAndDeallocate(rAllocator);
    delete pChildArr;
  }
  assert(0 <= m_nBCount && m_nBCount <= UPBLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
UIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::CutTailFromIndex
(
  UIndex     nUIndex,
  Allocator& rAllocator
)
{
  BIndex nBIndex;
  RecalcIndexes(nUIndex, nBIndex);

  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= nBIndex && nBIndex < m_nBCount);
  assert(NULL != m_arrItems[nBIndex]);

  BTreeArrayBlockInterfaceType* pChildArr = m_arrItems[nBIndex];
  const UIndex                  nULevel   = pChildArr->CutTailFromIndex(nUIndex, rAllocator) + 1;

  BTreeArrayBlockInterfaceType**                   ppArr = m_arrItems + nBIndex;
  const BTreeArrayBlockInterfaceType* const* const ppEnd = m_arrItems + m_nBCount;

  if ( 0 == pChildArr->GetItemsCount() )
    m_nBCount = nBIndex;
  else
  {
    ppArr++;
    m_nBCount = nBIndex + 1;
  }

  for(; ppEnd != ppArr; ppArr++)
  {
    (*ppArr)->DestroyAndDeallocate(rAllocator);
    delete *ppArr;
  }
  UndefineUCount();

  assert(NULL != m_pUpNode);
  assert(0 <= m_nBCount && m_nBCount <= UPBLOCK_SIZE);
  assert(0 < nULevel);
  return nULevel;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
bool BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::UniteWith
(
  BTreeArrayBlockInterfaceType* pRight,
  Allocator&                    //rAllocator
)
{
  assert(NULL != pRight);
  assert(this != pRight);
  assert(0 < pRight->GetItemsCount() && pRight->GetItemsCount() <= UPBLOCK_SIZE);
  assert(pRight->GetItemsCount() == static_cast<BTreeArrayUpBlock*>(pRight)->m_nBCount);
  assert(0 < m_nBCount && m_nBCount <= UPBLOCK_SIZE);
  assert(NULL != m_pUpNode);

  if ( UPBLOCK_SIZE <= size_t(m_nBCount) )
    return false;

  BIndex                   nMovCount   = UPBLOCK_SIZE - m_nBCount;
  const BIndex             nRCount     = pRight->GetItemsCount();
  BTreeArrayUpBlock* const pRightBlock = static_cast<BTreeArrayUpBlock*>(pRight);

  BTreeArrayBlockInterfaceType** pTo   = m_arrItems + m_nBCount;
  BTreeArrayBlockInterfaceType** pFrom = pRightBlock->m_arrItems;

  if ( nRCount <= nMovCount )
  {
    // all items of right block will be moved to this block

    for(const BTreeArrayBlockInterfaceType* const* const pEnd = pFrom + nRCount;
        pEnd != pFrom;
        pTo++, pFrom++)
    {
      assert(NULL != *pFrom);

      *pTo = *pFrom;
      (*pTo)->SetUpNode(this);
    }

    //-------------------------------------------------------------------indexes
    if ( m_pUpNode == pRightBlock->m_pUpNode )
    {
      if ( UINDEX_LIMIT != m_nUCount )
      {
        m_nUCount = UINDEX_LIMIT != pRightBlock->m_nUCount
                      ? m_nUCount + pRightBlock->m_nUCount
                      : UINDEX_LIMIT;
      }
    }
    else
    {
      UndefineUCount();
      pRightBlock->UndefineUCount();
    }

    m_nBCount += nRCount;
    pRightBlock->m_nBCount = 0;//for deleting pRight
    //--------------------------------------------------------------------------

    assert(m_nBCount <= UPBLOCK_SIZE);
    return true;
  }

  for(const BTreeArrayBlockInterfaceType* const* const pEnd = pFrom + nMovCount;
      pEnd != pFrom;
      pTo++, pFrom++)
  {
    assert(NULL != *pFrom);

    *pTo = *pFrom;
    (*pTo)->SetUpNode(this);
  }

  pTo = pRightBlock->m_arrItems;
  for(const BTreeArrayBlockInterfaceType* const* const pEnd = pTo + nRCount;
      pEnd != pFrom;
      pTo++, pFrom++)
  {
    *pTo = *pFrom;
  }

  //---------------------------------------------------------------------indexes
  assert(m_nBCount + nMovCount == UPBLOCK_SIZE);
  assert(0 < nRCount - nMovCount);

  m_nBCount              = UPBLOCK_SIZE;
  pRightBlock->m_nBCount = nRCount - nMovCount;

  if ( m_pUpNode == pRightBlock->m_pUpNode )
    pRightBlock->m_nUCount = m_nUCount = UINDEX_LIMIT;
  else
  {
    UndefineUCount();
    pRightBlock->UndefineUCount();
  }
  //----------------------------------------------------------------------------

  return false;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::RemoveUpper()
{
  assert(UPBLOCK_SIZE >= m_nBCount);

  if ( 1 == m_nBCount )
  {
    BTreeArrayBlockInterfaceType* pUp = RemoveLastItem();

    assert(0 == m_nBCount);

    delete this;
    return pUp->RemoveUpper();
  }
  
  return this;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
UIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetCount()
{
  assert(UPBLOCK_SIZE >= m_nBCount);

  if ( UINDEX_LIMIT == m_nUCount )
  {
    m_nUCount = 0;

    BTreeArrayBlockInterfaceType** pBlock = m_arrItems;
    for(
        const BTreeArrayBlockInterfaceType *const *const pEnd = m_arrItems + m_nBCount;
        pEnd != pBlock;
        pBlock++
       )
    {
      m_nUCount += (*pBlock)->GetCount();
    }
  }
  assert(NULL != m_pUpNode);

  return m_nUCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetItemsCount() const
{
  return m_nBCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
UIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Depth() const
{
  assert(0 < m_nBCount);
  return 1 + m_arrItems[0]->Depth();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::IncreaseUCount()
{
  if ( UINDEX_LIMIT != m_nUCount )
    m_nUCount++;

  assert(NULL != m_pUpNode);
  m_pUpNode->IncreaseUCount();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::IncreaseArrayCount()
{
  if ( UINDEX_LIMIT != m_nUCount )
    m_nUCount++;

  assert(NULL != m_pUpNode);
  m_pUpNode->IncreaseArrayCount();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DecreaseArrayCount()
{
  if ( UINDEX_LIMIT != m_nUCount )
  {
    assert(0 < m_nUCount);
    m_nUCount--;
  }

  assert(NULL != m_pUpNode);
  m_pUpNode->DecreaseArrayCount();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DecreaseUCount()
{
  if ( UINDEX_LIMIT != m_nUCount )
    m_nUCount--;

  assert(NULL != m_pUpNode);
  m_pUpNode->DecreaseUCount();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DecreaseUCount(UIndex nCount)
{
  if ( UINDEX_LIMIT != m_nUCount )
    m_nUCount -= nCount;

  assert(NULL != m_pUpNode);
  m_pUpNode->DecreaseUCount(nCount);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::UndefineUCount()
{
  assert(NULL != m_pUpNode);

  m_nUCount = UINDEX_LIMIT;
  m_pUpNode->UndefineUCount();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
UIndex BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetMyIndex(BTreeArrayBlockInterfaceType* pSubBlock)
{
  assert(NULL != pSubBlock);
  assert(NULL != m_pUpNode);

  UIndex nUIndex = m_pUpNode->GetMyIndex(this);

  for(
      BTreeArrayBlockInterfaceType** pBlock = m_arrItems;
      pSubBlock != *pBlock;
      pBlock++
     )
  {
    nUIndex += (*pBlock)->GetCount();
  }

  return nUIndex;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetPrev(BTreeArrayBlockInterfaceType* pItem)
{
  assert(NULL != pItem);
  assert(NULL != m_pUpNode);
  
  BIndex nBIndex = 0;

  for(; pItem != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);
  assert(nBIndex < m_nBCount);

  return 0 < nBIndex ? m_arrItems[nBIndex - 1]->Last() : m_pUpNode->GetPrev(this);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetNext(BTreeArrayBlockInterfaceType* pItem)
{
  assert(NULL != pItem);
  
  BIndex nBIndex = 0;

  for(; pItem != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);
  assert(nBIndex < m_nBCount);

  nBIndex++;

  if ( nBIndex >= m_nBCount )
    return m_pUpNode->GetNext(this);

  return m_arrItems[nBIndex]->Begin();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetNext
(
  BTreeArrayBlockInterfaceType* pItem, ptrdiff_t nOffset
)
{
  assert(NULL != pItem);
  
  BIndex nBIndex = 0;

  for(; pItem != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);
  assert(nBIndex < m_nBCount);

  if ( 0 <= nOffset )
  {
    UIndex nUSubCount;
    for(nBIndex++; nBIndex < m_nBCount; nBIndex++)
    {
      nUSubCount = m_arrItems[nBIndex]->GetCount();

#ifdef _MSC_VER
// suppress C4018 warning: '<' : signed/unsigned mismatch
// for the line "if ( nOffset < nUSubCount )"
#pragma warning( push )
#pragma warning( disable : 4018 )
#else 
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsign-compare"
#endif

      if ( nOffset < nUSubCount )// nOffset and nUSubCount are positive here
        return m_arrItems[nBIndex]->GetElementReference(static_cast<UIndex>(nOffset));

#ifdef _MSC_VER
#pragma warning( pop ) 
#else 
#pragma GCC diagnostic pop
#endif

      nOffset -= nUSubCount;
    }

    return m_pUpNode->GetNext(this, nOffset);
  }

  assert(nOffset < 0);

  while(nBIndex --> 0)
  {
    nOffset += m_arrItems[nBIndex]->GetCount();

    if ( 0 <= nOffset )
      return m_arrItems[nBIndex]->GetElementReference(static_cast<UIndex>(nOffset));
  }

  assert(nOffset < 0);
  return m_pUpNode->GetNext(this, nOffset);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetPrevSibling
( 
  BTreeArrayBlockInterfaceType* pBlock,
  UIndex                        nULevel
)
{
  assert(0 <= nULevel);
  assert(NULL != pBlock);
  assert(NULL != m_pUpNode);

  BIndex nBIndex = 0;

  for(; pBlock != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);
  assert(nBIndex < m_nBCount);

  if ( 0 == nBIndex )
    return m_pUpNode->GetPrevSibling(this, nULevel + 1);

  --nBIndex;

  if ( 0 == nULevel )
    return m_arrItems[nBIndex];

  return m_arrItems[nBIndex]->GetLastBlockOnLevel(nULevel);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetNextSibling
( 
  BTreeArrayBlockInterfaceType* pBlock,
  UIndex                        nULevel
)
{
  assert(0 <= nULevel);
  assert(NULL != pBlock);
  assert(NULL != m_pUpNode);

  BIndex nBIndex = 0;

  for(; pBlock != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);
  assert(nBIndex < m_nBCount);

  nBIndex++;

  if ( nBIndex >= m_nBCount )
    return m_pUpNode->GetNextSibling(this, nULevel + 1);

  if ( 0 == nULevel )
    return m_arrItems[nBIndex];

  return m_arrItems[nBIndex]->GetFirstBlockOnLevel(nULevel);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::DeleteItem
(
  BTreeArrayBlockInterfaceType* pItem
)
{
  assert(NULL != pItem);
  assert(NULL != m_pUpNode);
  assert(0 < m_nBCount);

  BIndex nBIndex = 0;

  for(; pItem != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);

  DeleteItem(nBIndex);//pItem is not valid after this
  if ( 0 == m_nBCount )
    m_pUpNode->DeleteItem(this);
  // this was deleted
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::InsertAfter
(
  BTreeArrayBlockInterfaceType* pAfterThis,
  BTreeArrayBlockInterfaceType* pNewBlock
)
{
  assert(NULL != pAfterThis);
  assert(NULL != pNewBlock);
  assert(0 < m_nBCount);
  assert(NULL != m_pUpNode);

  BIndex nBIndex = 0;
  for(; pAfterThis != m_arrItems[nBIndex]; nBIndex++)//find index
    assert(nBIndex < m_nBCount);

  nBIndex++;

  assert(0 < nBIndex && nBIndex <= m_nBCount);

  if ( !IsFull() )
  {
    InsertItem(nBIndex, pNewBlock); 
    return;
  }

  BTreeArrayBlockInterfaceType* pPrevBlock = m_pUpNode->GetPrevSibling(this, 0);

  if ( NULL != pPrevBlock )
  {
    BTreeArrayUpBlock* pPrev = static_cast<BTreeArrayUpBlock*>(pPrevBlock);
    if ( !pPrev->IsFull() )
    {
      assert(0 < nBIndex);
      UndefineUCount();//children count changed
      BTreeArrayBlockInterfaceType* pMoveNode = RemoveFirstInsertItem(nBIndex - 1, pNewBlock);// first removed, so the nBIndex decreased
      pPrev->AddItem(pMoveNode);
      pPrev->UndefineUCount();
      return;
    }
  }

  BTreeArrayBlockInterfaceType* pNextBlock = m_pUpNode->GetNextSibling(this, 0);

  if ( NULL != pNextBlock )
  {
    BTreeArrayUpBlock* pNext = static_cast<BTreeArrayUpBlock*>(pNextBlock);
    if ( !pNext->IsFull() )
    {
      UndefineUCount();//children count changed

      if ( UPBLOCK_SIZE != nBIndex )
      {
        BTreeArrayBlockInterfaceType* pMoveNode = RemoveLastItem();
        InsertItem(nBIndex, pNewBlock);
        assert(NULL != pMoveNode);
        pNewBlock = pMoveNode;
      }

      pNext->InsertItem(0, pNewBlock);
      pNext->UndefineUCount();//count changed
      return;
    }
  }

  m_nUCount = UINDEX_LIMIT;

  BTreeArrayUpBlock* pNewColl = new BTreeArrayUpBlock;
  pNewColl->GetHalveFrom(*this);

  assert(!IsFull());
  assert(!(pNewColl->IsFull()));

  if ( nBIndex <= m_nBCount )
    InsertItem(nBIndex, pNewBlock);
  else
    pNewColl->InsertItem(BIndex(nBIndex - m_nBCount), pNewBlock);

  assert(UINDEX_LIMIT == m_nUCount);          //child count changed
  assert(UINDEX_LIMIT == pNewColl->m_nUCount);//child count changed

  m_pUpNode->InsertAfter(this, pNewColl);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline Allocator& BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::GetAllocatorRef()
{
  assert(NULL != m_pUpNode);
  return m_pUpNode->GetAllocatorRef();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline bool BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::IsFull() const
{
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(0 <= m_nBCount);

  return UPBLOCK_SIZE == m_nBCount;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
inline BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::RemoveLastItem()
{
  assert(0 < m_nBCount);
  assert(UPBLOCK_SIZE >= m_nBCount);
  assert(NULL != m_arrItems[m_nBCount - 1]);

  m_nBCount--;
  return m_arrItems[m_nBCount];
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
void BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::Pack
(
  UIndex     nLevel,
  bool&      bSubBlockPacked,
  Allocator& rAllocator
)
{
  assert(0 < nLevel);
  assert(0 <= m_nBCount && m_nBCount <= UPBLOCK_SIZE);
  assert(NULL != m_pUpNode);

  // here we join the subblocks of previous level to free extra memory
  if ( m_nBCount < 2 )
    return;

  if ( 1 == nLevel )
  {
    // m_nUCount == sum of m_nBCounts of subblocks
    //
    // BLOCK_SIZE   == 7
    // UPBLOCK_SIZE == 10
    // m_nBCount    == 4
    // m_nUCount    == 3 + 4 + 5 + 2 == 14
    //
    //
    //  0 1 2 3 4 5 6     0 1 2 3 4 5 6     0 1 2 3 4 5 6     0 1 2 3 4 5 6
    // +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+ 
    // |x|x|x| | | | |   |x|x|x|x| | | |   |x|x|x|x|x| | |   |x|x| | | | | |   <== subblocks
    // +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+ 
    // \                 |                 |                /
    //   \               \                /               /
    //     \              \              /              /
    //       \             \            /             /
    //         \            \          /            /
    //           \           \        /           /
    //             \          \      /          /
    //               \         \    /         /
    //                 \        |  /        /
    //                   \      | |       /
    //                     \    | |     /
    //                       \  | |   /
    //                       +-+-+-+-+-+-+-+-+-+-+
    //                       |*|*|*|*| | | | | | |   <== this block
    //                       +-+-+-+-+-+-+-+-+-+-+
    //                        0 1 2 3 4 5 6 7 8 9
    //                        
    // capacity of subblocks = m_nBCount * BLOCK_SIZE == 4 * 7 == 28
    // if the half of the capacity is enought to keep all items, we call the Fusion(..) function.
    // so
    // if ( m_nUCount <= m_nBCount * BLOCK_SIZE / 2 ) call Fusion();
    // <=>
    // if ( 2 * m_nUCount <= m_nBCount * BLOCK_SIZE ) call Fusion();
    //
    // Result:
    //
    //  0 1 2 3 4 5 6     0 1 2 3 4 5 6 
    // +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+        BLOCK_SIZE   == 7
    // |x|x|x|x|x|x|x|   |x|x|x|x|x|x|x|        UPBLOCK_SIZE == 10
    // +-+-+-+-+-+-+-+   +-+-+-+-+-+-+-+        m_nBCount    == 2
    // \                 |                      m_nUCount    == 7 + 1 == 8
    //   \               \                                                     . 
    //     \              \                                                    . 
    //       \             \                                                   . 
    //         \            \                                                  . 
    //           \           \                                                 . 
    //             \          \                                                . 
    //               \         \                                               . 
    //                 \        |                                               
    //                   \      |                                               
    //                     \    |                                               
    //                       \  |                                               
    //                       +-+-+-+-+-+-+-+-+-+-+                              
    //                       |*|*| | | | | | | | |                              
    //                       +-+-+-+-+-+-+-+-+-+-+                              
    //                        0 1 2 3 4 5 6 7 8 9                               
    //
    UIndex nCount = GetCount();// recalcs m_nUCount
    if ( 2 * nCount <= UIndex(m_nBCount * BLOCK_SIZE) )
    {
      Fusion(rAllocator);
      bSubBlockPacked = true;
      assert(2 * size_t(nCount) > size_t(m_nBCount * BLOCK_SIZE) || 1 == m_nBCount);
    }
  }
  else
  {
    if ( bSubBlockPacked ) 
      bSubBlockPacked = PackSubblocks(rAllocator);
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
bool BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::PackSubblocks(Allocator& rAllocator)
{
  assert(1 < m_nBCount);

  UIndex nSubItems = 0;
  BTreeArrayBlockInterfaceType** pBlock = m_arrItems;
  for(
      const BTreeArrayBlockInterfaceType *const *const pEnd = m_arrItems + m_nBCount;
      pEnd != pBlock;
      pBlock++
     )
  {
    nSubItems += (*pBlock)->GetItemsCount();
  }

  // nSubItems == sum of m_nBCounts of subblocks
  const bool bPack = 2 * nSubItems <= UIndex(m_nBCount * UPBLOCK_SIZE);
  if ( bPack )
  {
    Fusion(rAllocator);
    assert(2 * size_t(nSubItems) > size_t(m_nBCount * BLOCK_SIZE) || 1 == m_nBCount);
  }

  return bPack;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, UIndex UINDEX_LIMIT, class Allocator>
BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArrayUpBlock<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, UINDEX_LIMIT, Allocator>::RemoveFirstInsertItem
(
  BIndex                        nBIndex,
  BTreeArrayBlockInterfaceType* pItem
)
{
  assert(NULL != pItem);
  assert(0 <= nBIndex && nBIndex < m_nBCount && m_nBCount <= UPBLOCK_SIZE);

  BTreeArrayBlockInterfaceType*  pMoveNode = m_arrItems[0];
  BTreeArrayBlockInterfaceType** pTo       = m_arrItems;
  BTreeArrayBlockInterfaceType** pFrom     = pTo + 1;
  BTreeArrayBlockInterfaceType** pEnd      = &m_arrItems[nBIndex];

  //                  pTo
  //                  | pFrom
  //                  | |
  // +-+-+-+-+-+-+-+-+-+-+-+-+
  // | | | | | | | | | | | | | 
  // +-+-+-+-+-+-+-+-+-+-+-+-+
  //                  ^
  //                  |
  //                  pEnd
  while(pTo != pEnd)
    *pTo++ = *pFrom++;

  *pTo = pItem;
  pItem->SetUpNode(this);
  return pMoveNode;
}


}// namespace Detail
}// namespace FastCpp
#endif//FastCpp_BTREEARRAYUPBLOCK_HPP
//** (END OF FILE  : BTreeArrayUpBlock.hpp) ************************************
