//*****************************************************************************
// FILE..........: BTreeArrayUpBlockInterface.hpp
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2013-06-21 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_BTREEARRAYUPBLOCKINTERFACE_HPP
#define FastCpp_BTREEARRAYUPBLOCKINTERFACE_HPP

//--------------------------------------------------------------- import std lib
#include <cstddef>
//------------------------------------------------------------------------------

namespace FastCpp
{
namespace Detail
{
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> class BTreeArrayBlockInterface;
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> class ElementReference;

//------------------------------------------------------------------------------
//| |       |                 |                 |                            |

template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
class BTreeArrayUpBlockInterface
{
  public:
    typedef BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayBlockInterfaceType;
    typedef ElementReference        <T, UIndex, BIndex, BLOCK_SIZE, Allocator> ElementReferenceType;
  public:
    virtual                  ~BTreeArrayUpBlockInterface()
                              {
                              }
    virtual void              IncreaseUCount    () = 0;
    virtual void              IncreaseArrayCount() = 0;
    virtual void              DecreaseArrayCount() = 0;
    virtual void              DecreaseUCount    () = 0;
    virtual void              DecreaseUCount    (UIndex nCount) = 0;
    virtual void              UndefineUCount    () = 0;

    virtual UIndex            GetMyIndex        (BTreeArrayBlockInterfaceType* pBlock) = 0;
    virtual ElementReferenceType
                              GetPrev           (BTreeArrayBlockInterfaceType* pBlock) = 0;
    virtual ElementReferenceType
                              GetNext           (BTreeArrayBlockInterfaceType* pBlock) = 0;
    virtual ElementReferenceType
                              GetNext           (
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  ptrdiff_t                     nOffset
                                                ) = 0;
    virtual BTreeArrayBlockInterfaceType*
                              GetPrevSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                ) = 0;
    virtual BTreeArrayBlockInterfaceType*
                              GetNextSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                ) = 0;
    virtual void              DeleteItem        (BTreeArrayBlockInterfaceType* pBlock) = 0;
    virtual void              InsertAfter       (
                                                  BTreeArrayBlockInterfaceType* pAfterThis,
                                                  BTreeArrayBlockInterfaceType* pNewBlock
                                                ) = 0;
    virtual Allocator&        GetAllocatorRef   () = 0;
};


//------------------------------------------------------------------------------
//------------------------------------------------- inline function declarations
//--------------------------------------------------------------- public methods
//------------------------------------------------------------ protected methods
//-------------------------------------------------------------- private methods
//------------------------------------------------------------- inline functions
}// namespace Detail
}// namespace FastCpp
#endif//FastCpp_BTREEARRAYUPBLOCKINTERFACE_HPP
//** (END OF FILE  : BTreeArrayUpBlockInterface.hpp) ***************************
