//******************************************************************************
// FILE..........: ElementReference.hpp
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2013-06-08 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_ELEMENTREFERENCE_HPP
#define FastCpp_ELEMENTREFERENCE_HPP

//--------------------------------------------------------------- import std lib
#include <assert.h>
//------------------------------------------------------------------------------

namespace FastCpp
{
namespace Detail
{
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> class BTreeArrayBlock;
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator> class BTreeArrayInterface;

//------------------------------------------------------------------------------
//| |       |                 |                 |                            |
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
class ElementReference
{
  private:
    typedef BTreeArrayBlock    <T, UIndex, BIndex, BLOCK_SIZE, Allocator>  BTreeArrayBlockType;
    typedef BTreeArrayInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>  BTreeArrayInterfaceType;
  public:
    inline                    ElementReference  ();
    inline                    ElementReference  (
                                                  BTreeArrayBlockType* pBlock,
                                                  BIndex               nBIndex
                                                );
    inline                    ElementReference  (
                                                  BTreeArrayInterfaceType* pArray,
                                                  BIndex                   nBIndex
                                                );
  public:
    inline  bool              operator ==       (const ElementReference& rObj) const;
  public:
            // m_nBIndex == BLOCK_SIZE                     => m_pArray
            // m_nBIndex == BLOCK_SIZE && NULL == m_pArray => invalid reference
            // m_nBIndex <  BLOCK_SIZE                     => m_pBlock
            union 
            {
                BTreeArrayBlock    <T, UIndex, BIndex, BLOCK_SIZE, Allocator>* m_pBlock;
                BTreeArrayInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* m_pArray;
            }                 
                              m_uRef;
            BIndex            m_nBIndex;
};

//------------------------------------------------------------------------------
//------------------------------------------------- inline function declarations
//--------------------------------------------------------------- public methods
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ElementReference()
  : m_nBIndex(BLOCK_SIZE)
{
  m_uRef.m_pArray = NULL;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ElementReference
(
  BTreeArrayBlockType* pBlock,
  BIndex               nBIndex
)
  : m_nBIndex(nBIndex)
{
  assert(NULL != pBlock);
  m_uRef.m_pBlock = pBlock;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::ElementReference
(
  BTreeArrayInterfaceType* pArray,
  BIndex                   nBIndex
)
  : m_nBIndex(nBIndex)
{
  assert(NULL != pArray);
  m_uRef.m_pArray = pArray;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline bool ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator == (const ElementReference& rObj) const
{
  return m_nBIndex == rObj.m_nBIndex
         &&
         (
           BLOCK_SIZE == m_nBIndex ? rObj.m_uRef.m_pArray == m_uRef.m_pArray
                                   : rObj.m_uRef.m_pBlock == m_uRef.m_pBlock
         );
}

//------------------------------------------------------------ protected methods
//-------------------------------------------------------------- private methods
//------------------------------------------------------------- inline functions
}// namespace Detail
}// namespace FastCpp
#endif//FastCpp_ELEMENTREFERENCE_HPP
//** (END OF FILE  : ElementReference.hpp) *************************************
