//*****************************************************************************
// FILE..........: IsIntegral.hpp
// DESCRIPTION...: IsIntegral metafunction
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2014-05-11 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_ISINTEGRAL_HPP
#define FastCpp_ISINTEGRAL_HPP

//------------------------------------------------------------------------------
#ifndef                 FastCpp_INTEGERTYPE_TAG_HPP
#include <FastCpp/Metafunctions/IntegerType_Tag.hpp>
#endif//                FastCpp_INTEGERTYPE_TAG_HPP
//------------------------------------------------------------------------------
namespace FastCpp
{
//------------------------------------------------------------------------------
// Integral types:
//
//   signed char
//   short int
//   int
//   long int
//   long long int
//   unsigned char
//   unsigned short int
//   unsigned int
//   unsigned long int
//   unsigned long long int
//   bool
//   char
//   wchar_t
//
//------------------------------------------------------------------------------
struct IntegralType_Tag    {};
struct NotIntegralType_Tag {};


template <typename T>
struct IsIntegral
{
  enum { is_Integral = 0 };
  typedef NotIntegralType_Tag Type;
};


//-----------------------------------------------------five signed integer types


template <>
struct IsIntegral<signed char>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<short int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<long int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<long long int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


//----------------------------------------------------five unsigned integer type


template <>
struct IsIntegral<unsigned char>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<unsigned short int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<unsigned int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<unsigned long int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<unsigned long long int>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


//-----------------------------------------------------------bool, char, wchar_t


template <>
struct IsIntegral<bool>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<char>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


template <>
struct IsIntegral<wchar_t>
{
  enum { is_Integral = 1 };
  typedef IntegralType_Tag Type;
};


//------------------------------------------------------------------------------
}// namespace FastCpp
#endif//FastCpp_ISINTEGRAL_HPP
//** (END OF FILE  : IsIntegral.hpp) *******************************************
